#include <sstream>
#include <math.h>
#include "jello_monster.h"

using namespace std;

BITMAP *jello_monster::pmonster = 0;
int jello_monster::instances = 0;

jello_monster::jello_monster (void) {
	//update shared information
	if (instances == 0) {
		//assume image is not loaded
		ex_main::logline ("loading the monster bitmap into memory");
		pmonster = load_pcx ("Jello_Master_1.pcx", 0);
		if (pmonster == NULL) {
			throw ex_main (__FILE__, __LINE__, "could not load the specified file");
		}
	}
	instances++;

	//sert instance information
	x = 200;
	y = 200;
	am_hit = false;
	am_dead = false;
}

jello_monster::~jello_monster (void) {
	instances--;
	if (instances == 0) {
		ex_main::logline ("removing the monster bitmap from memory");
		destroy_bitmap (pmonster);
		pmonster = 0;
	}
}

void jello_monster::draw (BITMAP *target, tins_game *game, int _x, int _y, int _w, int _h) {
	float s = game->get_scale (z);
	int sw = s * pmonster->w;
	int sh = s * pmonster->h;
	int sx, sy;
	game->logical_to_pixel (x, z, &sx, &sy);
        
	if (!am_hit) {
		masked_stretch_blit (pmonster, target, 0, 0, pmonster->w, pmonster->h, sx - sw / 2, sy - sh, sw, sh);
	}
	else if (!am_dead) {
		ellipsefill (target, sx, sy - sh / 2, sw / 2, sh / 2, 0xffff0000);
	}
}

void jello_monster::tick (unsigned int millis) {
	x -= millis / 1000.0;
	if (dead_countdown > 0) {
		dead_countdown -= millis;
	}
	if (am_hit && dead_countdown <= 0) {
		am_dead = true;
	}
}

void jello_monster::position (float *px, float *py, float *pz) {
	if (px != 0)
		*px = x;
	if (py != 0)
		*py = y;
	if (pz != 0)
		*pz = z;
}

void jello_monster::hit (void) {
        am_hit = true;
	dead_countdown = 150;
}

bool jello_monster::is_hit (void) {
	return am_hit;
}

bool jello_monster::is_dead (void) {
	return am_dead;
}

int jello_monster::get_width (void) {
	return pmonster->w;
}

int jello_monster::get_height (void) {
	return pmonster->h;
}

jello_monster* jello_monster::spawn (void) {
	ex_main::logline ("spawning a new jello monster");
	jello_monster *j = new jello_monster ();
	j->x = 1.5;
	j->z = (rand() % 10000) / 10000.0;
	return j;
}

jello_monster* jello_monster::spawn (float init_x, float minz, float maxz) {
	jello_monster *j = new jello_monster ();
	j->x = init_x;
	j->x = 1.5;
	j->z = minz + (rand() % 10000) / 10000.0 * (maxz - minz);
	return j;
}

jello_monster* jello_monster::spawn (float init_x, float init_z) {
	jello_monster *j = new jello_monster ();
	j->x = init_x;
	j->x = 1.5;
	j->z = init_z;
	return j;
}
