#include "molecule.h"
#include "ex_main.h"

int molecule::instances;
BITMAP *molecule::pmolecule;

molecule::molecule (void) {
	//update shared information
	if (instances == 0) {
		//assume image is not loaded
		ex_main::logline ("loading the monster bitmap into memory");
		pmolecule = load_pcx ("Molecule.pcx", 0);
		if (pmolecule == NULL) {
			throw ex_main (__FILE__, __LINE__, "could not load the specified file");
		}
	}
	instances++;
}

molecule::~molecule (void) {
	instances--;
	if (instances == 0) {
		ex_main::logline ("removing the monster bitmap from memory");
		destroy_bitmap (pmolecule);
		pmolecule = 0;
	}
}

void molecule::draw (BITMAP *target, tins_game *game, int _x, int _y, int _w, int _h) {
	float s = game->get_scale (z);
	int sw = s * pmolecule->w;
	int sh = s * pmolecule->h;
	int sx, sy;
	game->logical_to_pixel (x, z, &sx, &sy);
        
	masked_stretch_blit (pmolecule, target, 0, 0, pmolecule->w, pmolecule->h, sx - sw / 2, sy - sh, sw, sh);
}

void molecule::tick (unsigned int millis) {
	x -= millis / 1000.0;
}

void molecule::position (float *px, float *py, float *pz) {
	if (px != 0)
		*px = x;
	if (py != 0)
		*py = y;
	if (pz != 0)
		*pz = z;
}

void molecule::move_to (float nx, float nz) {
	x = nx;
	z = nz;
}

int molecule::get_width (void) {
	return pmolecule->w;
}

int molecule::get_height (void) {
	return pmolecule->h;
}

