#include <stdlib.h>
#include <vector>
#include <math.h>
#include "tins_game.h"

using namespace std;

tins_game::tins_game (int _xo, int _yo, int _width, int _height, formula *f) {
	try {
		xo = _xo;
		yo = _yo;
		width = _width;
		height = _height;
		num_monsters = 0;

		vx = 0;

		pgrnd = create_bitmap (gndw, gndh);
		if (pgrnd == 0) {
			throw ex_main (__FILE__, __LINE__, "could not create ground bitmap");
		}
		for (int j = 0; j < gndh; j++) {
			for (int i = 0; i < gndw; i++) {
				unsigned long basecolor = 0xffbf4f4f;
				unsigned long redoffset = (rand() % 0x40) << 16; 
				unsigned long greenoffset = (rand() % 0x20) << 8;
				unsigned long blueoffset = (rand() % 0x20);
				putpixel (pgrnd, i, j, basecolor + redoffset + greenoffset + blueoffset);
			}
		}

		gen_mountains (width, height);

		phero = new hero ();
		num_shots = 0;
		gun_cooldown = 0;

		num_molecules = 0;

		lose = false;
		win = false;

		form = f;
	}
	catch (ex_main e) {
		ex_main::logline ("  |-> in tins_game::tins_game()");
		throw;
	}
}

tins_game::~tins_game (void) {
	try {
		destroy_bitmap (pmtn);
		destroy_bitmap (pgrnd);

		delete phero;

		for (int i = 0; i < num_monsters; i++) {
			delete monsters[i];
		}

	}
	catch (ex_main e) {
		ex_main::logline ("  |-> in tins_game::tins_game()");
		throw;
	}
}


void tins_game::restart (void) {
	for (int i = 0; i < num_monsters; i++) {
		delete monsters[i];
	}
	num_monsters = 0;

	vx = 0;

	num_shots = 0;
	gun_cooldown = 0;

	for (int i = 0; i < num_molecules; i++) {
		delete molecules[i];
	}
	num_molecules = 0;

	lose = false;
	win = false;

	form->clear();
}

void tins_game::draw (BITMAP *target) {
	try {
		//clear background
		clear_to_color (target, 0xffcfaf20);

		//draw the mountains
		float mo = vx / 2;
		while (mo >= width) {
			mo -= width;
		}
		blit (pmtn, target, 0, 0, -mo,        yo + height / 2 - 3 * height / 8, width, 3 * height / 16);
		blit (pmtn, target, 0, 0, width - mo, yo + height / 2 - 3 * height / 8, width, 3 * height / 16);

		//draw ground
		float texo = vx / width * gndw;
		while (texo >= gndw) {
			texo -= gndw;
		}
		V3D_f v1 = { xo - width / 2,         yo + height,     1,   texo,        0,    0xffffff00 };
		V3D_f v2 = { xo + width + width / 2, yo + height,     1,   texo + gndw, 0,    0xffffff00 };
		V3D_f v3 = { xo + width,             yo + height / 4, 2.5, texo + gndw, gndh, 0xffffff00 };
		V3D_f v4 = { xo,                     yo + height / 4, 2.5, texo,        gndh, 0xffffff00 };
		quad3d_f (target, POLYTYPE_PTEX, pgrnd, &v1, &v2, &v3, &v4);

		//sort everything
		sort_monsters ();
		sort_molecules ();

		//draw molecules
		for (int i = 0; i < num_molecules; i++) {
			molecules[i]->draw (target, this, xo, yo, width, height);
		}
		
		//draw shots
		for (int i = 0; i < num_shots; i++) {
			float shot_width = 0.05;
			float s = get_scale(shots[i].z);
			int x1, x2, y;
			logical_to_pixel (shots[i].x - shot_width / 2.0, shots[i].z, &x1, &y);
			logical_to_pixel (shots[i].x + shot_width / 2.0, shots[i].z, &x2, 0);

			hline (target, x1, y - 75 * s, x2, 0xffffffff);
		}

		//draw monsters
		for (int i = 0; i < num_monsters; i++) {
			monsters[i]->draw (target, this, xo, yo, width, height);
		}

		//draw hero
		phero->draw (target, this, xo, yo, width, height);

		char num [32];
		sprintf (num, "Monsters: %d", num_monsters);
		textprintf (target, font, 400, 0, 0xffffffff, num);

		if (lose) {
			textprintf_centre (target, font, width / 2, height / 4, 0xffffffff, "You Lose");
			draw_button (target, 450, 420, 100, 20);

		}
		else if (win) {
			textprintf_centre (target, font, width / 2, height / 4, 0xffffffff, "You Win");
			draw_button (target, 450, 420, 100, 20);
		}
	}
	catch (ex_main e) {
		ex_main::logline ("  |-> in tins_game::draw()");
		throw;
	}
}

void tins_game::tick (unsigned int millis) {
	try {
		if (lose || win) {
			if (mouse_x >= 450 && mouse_x <= 550 && mouse_y >= 420 && mouse_y <= 440) {
				if (mouse_b & 1) {
					restart ();
				}
			}
			return;
		}
		vx += width * millis / 2000.0;

		//move molecules
		for (int i = 0; i < num_molecules; i++) {
			molecules[i]->tick (millis);
		}
		
		//check to see if we should remove any molecules
		for (int i = 0; i < num_molecules; i++) {
			float x;
			molecules[i]->position (&x, 0, 0);
			if (x < -0.5) {
				delete molecules[i];
				molecules[i] = molecules[num_molecules - 1];
				num_molecules--;
				i--;
			}
		}

		//check to see if we intersect any molecules
		for (int m = 0; m < num_molecules; m++) {
			float molx, molz;
			int molpx, molpy;
			int molx1, moly1, molx2, moly2;

			float hx, hz;
			int hpx, hpy;
			int hx1, hy1, hx2, hy2;
			
			molecules[m]->position (&molx, 0, &molz);
			float ms = get_scale (molz);
			logical_to_pixel (molx, molz, &molpx, &molpy);
			molx1 = molpx - ms * molecules[m]->get_width() / 2;
			moly1 = molpy - ms * molecules[m]->get_height();
			molx2 = molpx + ms * molecules[m]->get_width() / 2;
			moly2 = molpy;

			phero->position (&hx, 0, &hz);
			float hs = get_scale (hz);
			logical_to_pixel (hx, hz, &hpx, &hpy);
			hx1 = hpx - hs * phero->get_width() / 2;
			hy1 = hpy - hs * phero->get_height();
			hx2 = hpx + hs * phero->get_width() / 2;
			hy2 = hpy;

			if (spr_intersect (molx1, moly1, molx2, moly2, hx1, hy1, hx2, hy2)) {
				//we have a collision
				delete molecules[m];
				molecules[m] = molecules[num_molecules - 1];;
				num_molecules--;
				i--;
                                form->add_part ();				
				if (form->got_all_parts()) {
					win = true;
					return;
				}
			}
			
		}
		
		//check to see if we intersect any monsters
		for (int m = 0; m < num_monsters; m++) {
			float monx, monz;
			int monpx, monpy;
			int monx1, mony1, monx2, mony2;

			float hx, hz;
			int hpx, hpy;
			int hx1, hy1, hx2, hy2;
			
			monsters[m]->position (&monx, 0, &monz);
			float ms = get_scale (monz);
			logical_to_pixel (monx, monz, &monpx, &monpy);
			monx1 = monpx - ms * monsters[m]->get_width() / 2;
			mony1 = monpy - ms * monsters[m]->get_height();
			monx2 = monpx + ms * monsters[m]->get_width() / 2;
			mony2 = monpy;

			phero->position (&hx, 0, &hz);
			float hs = get_scale (hz);
			logical_to_pixel (hx, hz, &hpx, &hpy);
			hx1 = hpx - hs * phero->get_width() / 2;
			hy1 = hpy - hs * phero->get_height();
			hx2 = hpx + hs * phero->get_width() / 2;
			hy2 = hpy;

			if (spr_intersect (monx1, mony1, monx2, mony2, hx1, hy1, hx2, hy2)) {
				//we have a collision
				lose = true;
				return;
			}
			
		}
		//check to see if we should remove any monsters
		for (int i = 0; i < num_monsters; i++) {
			float x;
			monsters[i]->position (&x, 0, 0);
			if (x < -0.5 || monsters[i]->is_dead ()) {
				float mx, mz;
				monsters[i]->position (&mx, 0, &mz);
				create_molecule_maybe (mx, mz);
				delete monsters[i];
				monsters[i] = monsters[num_monsters - 1];
				monsters[num_monsters - 1] = 0;
				num_monsters--;
				i--;
			}
		}

		//now check to see if we should spawn a new monster
		if (num_monsters < max_monsters) {
			float prob = (millis / 400.0) * (float)(max_monsters - num_monsters) / (float)max_monsters;
			float num = (rand() % 10001) / (10000.0);
			if (num <= prob) {
				monsters[num_monsters] = jello_monster::spawn ();
				num_monsters++;
			}
		}

		//now check for projectiles
		if (gun_cooldown > 0)
			gun_cooldown -= millis;

		if (key[KEY_SPACE] && gun_cooldown <= 0) {
			ex_main::logline ("creating shot");
			float z;
			phero->position (0, 0, &z);
			create_shot (z);
		}

		//now update our shots
		for (int i = 0; i < num_shots; i++) {
			int shotx1, shotx2, shoty;
			logical_to_pixel (shots[i].x, shots[i].z, &shotx1, &shoty);
			logical_to_pixel (shots[i].x + 0.05, shots[i].z, &shotx2, 0);

			float ss = get_scale (shots[i].z);
			shoty -= ss * 75;

			for (int m = 0; m < num_monsters; m++) {
				int monsterx1, monstery1, monsterx2, monstery2;
				float monsterx, monsterz;
				int monsterpx, monsterpy;
				
				monsters[m]->position (&monsterx, 0, &monsterz);
				float ms = get_scale (monsterz);
				logical_to_pixel (monsterx, monsterz, &monsterpx, &monsterpy);
				monsterx1 = monsterpx - ms * monsters[m]->get_width() / 2;
				monstery1 = monsterpy - ms * monsters[m]->get_height();
				monsterx2 = monsterpx + ms * monsters[m]->get_width() / 2;
				monstery2 = monsterpy;

				if (shoty >= monstery1 && shoty <= monstery2) {
					if (shotx1 < monsterx1 && shotx2 > monsterx1) {
						//we have a collision
						if (!monsters[m]->is_hit()) {
							monsters[m]->hit();
							shots[i] = shots[num_shots - 1];
							num_shots--;
							i--;
						}
					}
				}
				
			}
			for (int i = 0; i < num_shots; i++) {
				shots[i].x += millis * 0.002;
				if (shots[i].x >= 1.5) {
					ex_main::logline ("removing shot");
					shots[i] = shots[num_shots - 1];
					num_shots--;
					i--;
				}
			}
		}
		
		//now update all monsters
		for (int i = 0; i < num_monsters; i++) {
			monsters[i]->tick (millis);
		}
		phero->tick (millis);
	}
	catch (ex_main e) {
		ex_main::logline ("  |-> in tins_game::tick()");
		throw;
	}
}

float tins_game::get_scale (float z) {
	return 1.0 / (5.0 * z + 1.0);
}

void tins_game::logical_to_pixel (float x, float z, int *px, int *py) {
	float s = 5.0 / (6.0 * (z + 1.0)) + 1.0 / 6.0;

	if (py != 0)
		*py = yo + 0.25 * height + (0.75 * height) * (5.0 * s - 2.0) / 3.0;
	if (px != 0)
		*px = xo + ((2.0 * x - 1.0) * s + 1) / 2.0 * width;
}

void tins_game::gen_mountains (int width, int height) {
	try {
		pmtn = create_bitmap (width, 3 * height / 16);
		if (pmtn == 0) {
			throw ex_main (__FILE__, __LINE__, "could not create mountain bitmap");
		}
		clear_to_color (pmtn, 0xffcfaf20);

		int skyh = rand() % (height / 8);
		int *heights = new int[width];
		for (int i = 0; i < width; i++) {
			skyh += rand() % 5 - 2;
			if (skyh < 0)
				skyh = 0;
			else if (skyh > height / 8)
				skyh = height / 8;
			heights[i] = skyh;
		}

		for (int i = 0; i < width; i++) {
			int skyh = (heights[i] + heights[width - i - 1]) / 2;
			for (int j = skyh; j < 3 * height / 16; j++) {
				unsigned long basecolor = 0xffbf4f4f;
				unsigned long redoffset = (rand() % 0x40) << 16; 
				unsigned long greenoffset = (rand() % 0x20) << 8;
				unsigned long blueoffset = (rand() % 0x20);
				putpixel (pmtn, i, j, basecolor + redoffset + greenoffset + blueoffset);
			}
			
		}
		delete[] heights;
	}
	catch (ex_main e) {
		ex_main::logline ("  |-> in tins_game::gen_mountains()");
	}
}

void tins_game::create_shot (float z) {
	if (num_shots < max_shots) {
		shots[num_shots].x = 0.15;
		shots[num_shots].z = z;
		num_shots++;
		gun_cooldown = 100;
	}
}

void tins_game::create_molecule_maybe (float x, float z) {
	if (num_molecules < max_molecules) {
		float p = 0.1;
		float roll = (rand() % 10001) / 10000.0;
		if (roll <= p) {
			molecules[num_molecules] = new molecule ();
			molecules[num_molecules]->move_to (x, z);
			num_molecules++;
		}
	}
}

bool tins_game::spr_intersect (int px1, int py1, int px2, int py2, int qx1, int qy1, int qx2, int qy2) {
	int x;
	int y;

	x = px1;
	y = py1;
	if (x >= qx1 && x <= qx2 && y >= qy1 && y <= qy2) {
		return true;
	}
	x = px2;
	y = py1;
	if (x >= qx1 && x <= qx2 && y >= qy1 && y <= qy2) {
		return true;
	}
	x = px1;
	y = py2;
	if (x >= qx1 && x <= qx2 && y >= qy1 && y <= qy2) {
		return true;
	}
	x = px2;
	y = py2;
	if (x >= qx1 && x <= qx2 && y >= qy1 && y <= qy2) {
		return true;
	}
	return false;
}

void tins_game::sort_molecules () {
	//crappy bubble sort
	for (int i = 0; i < num_molecules; i++) {
		float iz;
		molecules[i]->position (0, 0, &iz);
		for (int j = i; j < num_molecules; j++) {
                        float jz;
			molecules[j]->position (0, 0, &jz);
			if (iz < jz) {
				molecule *temp;
				temp = molecules[i];
				molecules[i] = molecules[j];
				molecules[j] = temp;
			}
		}
	}
}

void tins_game::sort_monsters () {
	//crappy bubble sort
	for (int i = 0; i < num_monsters; i++) {
		float iz;
		monsters[i]->position (0, 0, &iz);
		for (int j = i; j < num_monsters; j++) {
                        float jz;
			monsters[j]->position (0, 0, &jz);
			if (iz < jz) {
				jello_monster *temp;
				temp = monsters[i];
				monsters[i] = monsters[j];
				monsters[j] = temp;
			}
		}
	}
}

void tins_game::draw_button (BITMAP *target, int x, int y, int w, int h) {
	int realwidth = max (7 * 8, w);
	int realheight = max (2 * 8, h);

	rectfill (target, x, y, x + w, y + h, 0xff000000);
	rect (target, x, y, x + w, y + h, 0xffffff00);
	textprintf_centre (target, font, x + w / 2, y + h / 4, 0xffffffff, "Retry");
}
